//****************************************************************************
//**
//**    SYS_MAIN.CPP
//**    System Main
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for SYSW_Argc, SYSW_Argv,
                               // and SYSW_MessageBox
#include "sys_defs.h"
#include "sys_main.h"
//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SAFEMALLOC_MARKER 0x2b3d1234
#define SAFEMALLOC_USEMARKER // enable malloc/free marker verify

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static float sys_baseTime;
static float sys_lastTime;

static char sys_conCommandLine[256];

//============================================================================
//    INTERFACE DATA
//============================================================================
char sys_programPath[SYS_MAX_PATH];
float sys_curTime, sys_frameTime;

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void UpdateFrameTimes();
static void ResetFrameTimes();

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void UpdateFrameTimes()
{
	sys_curTime = M_TimeFloat() - sys_baseTime;
	sys_frameTime = sys_curTime - sys_lastTime;
	sys_lastTime = sys_curTime;
}

static void ResetFrameTimes()
{
	sys_baseTime = M_TimeFloat();
	sys_lastTime = sys_curTime = sys_frameTime = 0.0;
}

static bbool ConsoleReceiver(in_event_t* event)
{
    char tempKeyName[2] = {0,0};

    if (event->eventType != INEV_PRESS)
        return(1);
    switch(event->key)
    {
    case INKEY_ENTER:
        CON_Printf("\b \b\n", sys_conCommandLine);
        CON_Execute(false, sys_conCommandLine);
        sys_conCommandLine[0] = 0;
        break;
    case INKEY_LEFTARROW:
    case INKEY_BACKSPACE:
        SYS_Debugf("\b \b");
        if (sys_conCommandLine[0])
            sys_conCommandLine[strlen(sys_conCommandLine)-1] = 0;
        break;
    case INKEY_SPACE:
        strcat(sys_conCommandLine, " ");
        break;
    default:
        if ((event->key < 32) || (event->key > 127))
            break;
        // single-character key
        tempKeyName[0] = event->key;
        if (event->flags & INEVF_SHIFT)
            tempKeyName[0] = IN_GetShiftedKey(event->key);
        strcat(sys_conCommandLine, tempKeyName);
        break;
    }
    return(1);
}

static void PrintCallback(char* str)
{
    SYS_Debugf(str);
}

CON_FUNC(Quit, 0)
{
    SYS_Quit();
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// The "Big Three" control functions

// initialization
void SYS_Init()
{
    MEM_Init(); // initialize memory system
	XF_Init(); // initialize file system
    RES_Init(); // initialize resource manager
	IN_Init(); // initialize user input
    
	SYS_InitDebug();

    CON_SetPrintCallback(PrintCallback); // this can be safely called before CON_Init
    CON_Init(); // initialize command console

    // console test
    IN_AddReceiver(ConsoleReceiver);

	if (!M_CmdLineOption("noalert", 0)) // example of command line option use
	{
		SYS_Alertf("Hi!  I'm a small test of the internal command console.\n\n"
                   "Alt-tab over to the Debugging window and move it to where "
                   "you can see it, then select the main window again.\n\n"
                   "After that, type LISTCMDS to see a list of valid console "
                   "commands.  To exit the program, use the QUIT command.\n\n"
				   "BTW, you can run this program with the -noalert "
				   "option to skip this message.");
	}

	ResetFrameTimes(); // initialize frame times from this point
}

// main loop
void SYS_Frame()
{
	UpdateFrameTimes();
	
	RES_Frame(); // update resources
    IN_Frame(); // process user input

    // console test
    SYS_Debugf("\r>%s_", sys_conCommandLine);
}

// de-initialization
void SYS_Shutdown()
{	
    static bbool shuttingDown = 0;
    if (shuttingDown)
        return; // recursive shutdown block
    shuttingDown = 1;

	SYS_ShutdownDebug();

	CON_Shutdown();
    IN_Shutdown();
    RES_Shutdown();
    XF_Shutdown();
    MEM_Shutdown();
}

// alert notification
void SYS_Alertf(char* text, ...)
{
	static char textBuffer[2048];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox(SYS_APPTITLE " Notification", MB_OK, textBuffer);
}

// initialize debugging window
void SYS_InitDebug()
{
	SYSW_InitDebug();
}

// shut down debugging window
void SYS_ShutdownDebug()
{
	SYSW_ShutdownDebug();
}

// debugging window output
void SYS_Debugf(char* text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_DebugString(textBuffer);
}

// exit without error
void SYS_Quit()
{
	SYS_Shutdown();
	exit(0);
}

// exit with fatal error
void SYS_Errorf(char *text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox("Fatal Error", MB_OK, textBuffer);
	SYS_Shutdown();
	exit(1);
}

// command line argument functions, call windows equivalents
byte SYS_Argc()
{
	return(SYSW_Argc());
}

char* SYS_Argv(byte index)
{
	return(SYSW_Argv(index));
}

// safe malloc and free wrappers, with optional marker verify (4 byte cost)
void* SYS_SafeMalloc(int size)
{
#ifdef SAFEMALLOC_USEMARKER	
	size += 4;
#endif
	void *res = malloc(size);
	if (!res)
		SYS_Errorf("SYS_SafeMalloc: Out of memory on %d bytes", size);
#ifdef SAFEMALLOC_USEMARKER	
	*((long *)res) = SAFEMALLOC_MARKER;
	return((long *)res+1);
#else
	return(res);
#endif
}

void SYS_SafeFree(void **ptr)
{
	long *p;
	if (!(*ptr))
		SYS_Errorf("SYS_SafeFree: Attempt to free null pointer");
	p = (long *)*ptr;
#ifdef SAFEMALLOC_USEMARKER	
	p -= 1;
	if (*p != SAFEMALLOC_MARKER)
    {
		if (*p == (SAFEMALLOC_MARKER ^ 0xFFFFFFFF))
            SYS_Errorf("SYS_SafeFree: Attempt to free pointer twice");
        else
            SYS_Errorf("SYS_SafeFree: Attempt to free non-SafeMalloc'ed pointer");
    }
    *p = SAFEMALLOC_MARKER ^ 0xFFFFFFFF;
#endif
	free(p);
	*ptr = NULL;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SYS_MAIN.CPP
//**
//****************************************************************************

